from rich.columns import Columns

from dataclasses import dataclass
from enum import StrEnum
from functools import cached_property
from typing import Iterable

import funcy
from rdflib import URIRef, Literal
from textual.widget import Widget
from textual.widgets import Label, Static, ListView, ListItem

from iolanta.facets.facet import Facet
from iolanta.models import NotLiteralNode


class TermStatus(StrEnum):
    STABLE = 'stable'
    ARCHAIC = 'archaic'
    TESTING = 'testing'
    UNSTABLE = 'unstable'


@dataclass
class TermAndStatus:
    term: URIRef
    status: TermStatus


class TermsContent(Static):
    DEFAULT_CSS = '''
    TermsContent {
        padding: 1;
    }
    '''


class OntologyFacet(Facet[Widget]):
    @cached_property
    def grouped_terms(self) -> dict[NotLiteralNode | None, list[TermAndStatus]]:
        rows = self.stored_query('terms.sparql', iri=self.iri)
        grouped = [
            (
                row.get('group'),
                TermAndStatus(
                    term=row['term'],
                    status=status,
                ),
            )
            for row in rows
            if (status := TermStatus(
                (status_literal := row.get('status'))
                and status_literal.value
                or 'stable'
            )) != TermStatus.ARCHAIC
        ]

        return funcy.group_values(grouped)

    def _stream_columns(self) -> Iterable[str]:
        for group, rows in self.grouped_terms.items():
            group_title = self.render(
                group,
                environments=[URIRef('https://iolanta.tech/env/title')],
            ) if group is not None else '<Ungrouped>'

            content = '\n'.join([
                self.render(
                    row.term,
                    environments=[
                        URIRef('https://iolanta.tech/cli/link'),
                    ],
                )
                for row in rows
            ])

            column = f'[b]{group_title}[/b]\n\n' + content

            yield column

    def show(self) -> Widget:
        return TermsContent(Columns(self._stream_columns(), padding=(1, 2)))



from pathlib import Path
from typing import Iterator, TypedDict
from urllib.parse import unquote

from rdflib import URIRef
from yarl import URL

from iolanta.models import Quad, Triple


def triples_to_quads(
    triples: Iterator[Triple],
    graph: URIRef,
) -> Iterator[Quad]:
    """Convert sequence of triples to sequence of quads."""
    yield from (
        triple.as_quad(graph)
        for triple in triples
    )


Doc = TypedDict(
    'Doc', {
        '__doc__': str,
    },
)


def doc(docstring: str) -> Doc:
    """Document a dataclass field."""
    return {
        '__doc__': docstring,
    }


def construct_metadata_graph_name(iri: URIRef) -> URIRef:
    """Graph to store meta information about the file or directory."""
    return URIRef(f'{iri}/_metadata')


def url_to_iri(url: URL) -> URIRef:
    return URIRef(str(url))


def path_to_url(path: Path) -> URL:
    """Construct a file:// URL."""
    return URL(f'file://{path}')


def path_to_iri(path: Path) -> URIRef:
    """Construct a file:// IRI."""
    return URIRef(f'file://{path}')


def url_to_path(url: URL) -> Path:
    """Convert a file:// URL into a UNIX path."""
    return Path(unquote(url.path))



import traceback
from dataclasses import dataclass
from typing import ItemsView, Iterable, Mapping, Any

from boltons.iterutils import remap, default_enter
from rdflib import (
    URIRef, Variable, RDF, ConjunctiveGraph, Graph, RDFS, FOAF,
    Namespace, OWL, DC, VANN,
)
from rdflib.plugins.sparql.algebra import translateQuery
from rdflib.plugins.sparql.evaluate import evalQuery
from rdflib.plugins.sparql.parser import parseQuery
from rdflib.plugins.sparql.parserutils import CompValue
from rdflib.plugins.sparql.sparql import Query
from rdflib.query import Processor
from rdflib.term import Node

from iolanta.models import Triple, TripleWithVariables


def construct_flat_triples(algebra: Mapping[str, Any]) -> Iterable[Triple]:
    if isinstance(algebra, Mapping):
        for key, value in algebra.items():
            if key == 'triples':
                yield from [Triple(*raw_triple) for raw_triple in value]

            else:
                yield from construct_flat_triples(value)


@dataclass
class GlobalSPARQLProcessor(Processor):
    graph: ConjunctiveGraph

    def _download_namespace(self, namespace: Namespace):
        if namespace == VANN:
            iri = URIRef('https://vocab.org/vann/vann-vocab-20100607.rdf')
        else:
            iri = URIRef(namespace)

        try:
            self.graph.get_graph(iri)
        except IndexError:
            print(f'DOWNLOADING {namespace}')
            self.graph.get_context(iri).parse(iri)
            print(f'DOWNLOADED {namespace}!')

    def query(
        self,
        strOrQuery,
        initBindings={},
        initNs={},
        base=None,
        DEBUG=False,
    ):
        """
        Evaluate a query with the given initial bindings, and initial
        namespaces. The given base is used to resolve relative URIs in
        the query and will be overridden by any BASE given in the query.
        """
        if not isinstance(strOrQuery, Query):
            parsetree = parseQuery(strOrQuery)
            query = translateQuery(parsetree, base, initNs)

            triples = construct_flat_triples(query.algebra)
            for triple in triples:
                self.load_data_for_triple(triple, bindings=initBindings)
        else:
            query = strOrQuery
        return evalQuery(self.graph, query, initBindings, base)

    def load_data_for_triple(
        self,
        triple: TripleWithVariables,
        bindings: dict[str, Node],
    ):
        """Load data for a given triple."""
        triple = TripleWithVariables(*[
            self.resolve_term(term, bindings=bindings)
            for term in triple
        ])

        subject, *_etc = triple

        if isinstance(subject, URIRef):
            namespaces = [RDF, RDFS, OWL, FOAF, DC, VANN]

            for namespace in namespaces:
                if subject == URIRef(namespace) or subject in namespace:
                    self._download_namespace(namespace)

    def resolve_term(self, term: Node, bindings: dict[str, Node]):
        """Resolve triple elements against initial variable bindings."""
        if isinstance(term, Variable):
            return bindings.get(
                str(term),
                term,
            )

        return term



import logging
from pathlib import Path
from typing import Optional

import typer
from documented import DocumentedError
from rich.console import Console
from rich.logging import RichHandler
from rich.markdown import Markdown
from rich.table import Table
from typer import Argument, Context, Option, Typer

from iolanta.cli.formatters.choose import cli_print
from iolanta.cli.models import LogLevel
from iolanta.iolanta import Iolanta
from iolanta.models import QueryResultsFormat

logger = logging.getLogger('iolanta')


def construct_app() -> Typer:
    iolanta = Iolanta(logger=logger)

    cli = Typer(
        no_args_is_help=True,
        context_settings={
            'obj': iolanta,
        },
    )

    plugins = iolanta.plugins
    for plugin in plugins:
        if (subcommand := plugin.typer_app) is not None:
            cli.add_typer(subcommand)

    return cli


app = construct_app()


@app.callback()
def callback(
    ctx: Context,
    log_level: LogLevel = LogLevel.ERROR,
):
    """Iolanta Linked Data browser."""
    level = {
        LogLevel.DEBUG: logging.DEBUG,
        LogLevel.INFO: logging.INFO,
        LogLevel.WARNING: logging.WARNING,
        LogLevel.ERROR: logging.ERROR,
    }[log_level]

    iolanta: Iolanta = ctx.obj
    iolanta.logger.level = level

    logging.basicConfig(
        level=level,
        format='%(message)s',
        datefmt="[%X]",
        handlers=[RichHandler()],
        force=True,
    )


@app.command(name='browse')
def render_command(
    context: Context,
    url: str,
    environment: str = Option(
        'https://iolanta.tech/cli/interactive',
        '--as',
    ),
    print_stack: bool = Option(False, '--stack'),
):
    """Render a given URL."""
    iolanta: Iolanta = context.obj

    node = iolanta.string_to_node(url)

    try:
        renderable, stack = iolanta.render(
            node=node,
            environments=[
                iolanta.string_to_node(environment),
            ],
        )

    except DocumentedError as documented_error:
        if iolanta.logger.level == logging.DEBUG:
            raise

        Console().print(
            Markdown(
                str(documented_error),
                justify='left',
            ),
        )
        raise typer.Exit(1)

    except Exception as err:
        if iolanta.logger.level == logging.DEBUG:
            raise

        Console().print(str(err))
        raise typer.Exit(1)

    else:
        console = Console()

        if print_stack:
            console.print(stack)

        Console().print(renderable)


@app.command()
def namespaces(
    context: Context,
):
    """Registered namespaces."""
    iolanta: Iolanta = context.obj

    table = Table(
        'Namespace',
        'URL',
        show_header=True,
        header_style='bold magenta',
    )

    for namespace, url in iolanta.graph.namespaces():   # type: ignore
        table.add_row(namespace, url)

    Console().print(table)


@app.command()
def query(
    context: Context,
    fmt: QueryResultsFormat = Option(
        default=QueryResultsFormat.PRETTY,
        metavar='format',
    ),
    query_text: Optional[str] = Argument(
        None,
        metavar='query',
        help='SPARQL query text. Will be read from stdin if empty.',
    ),
    use_qnames: bool = Option(
        default=True,
        help='Collapse URLs into QNames.',
    ),
):
    """Query Iolanta graph with SPARQL."""
    iolanta: Iolanta = context.obj

    cli_print(
        query_result=iolanta.query(query_text),
        output_format=fmt,
        display_iri_as_qname=use_qnames,
        graph=iolanta.graph,
    )



import textwrap
from dataclasses import dataclass, field
from typing import List

from documented import DocumentedError
from rdflib import Literal, URIRef
from rdflib.term import Node

from iolanta.models import NotLiteralNode


@dataclass
class PageNotFound(DocumentedError):
    """
    Page not found by IRI: `{self.iri}`.

    !!! error "Page not found by IRI: `{self.iri}`"
        Every page on your documentation site, just as any other entity described on
        it, has an IRI — a unique identifier similar to a URL. IRI can be generated
        in two alternative ways.

          1. If you wanted to use IRI generated automatically, then please confirm
             that in the `docs` directory there is a file under the following path:

             ```
             {self.possible_path}
             ```

          2. If you wanted to use IRI that you specified explicitly then confirm
             there is a Markdown file somewhere in `docs` directory which contains
             the following text in its header:

             ```markdown
             ---
             $id: {self.possible_id}
             ---
             ```
    """

    iri: URIRef

    @property
    def possible_path(self) -> str:
        """Guess on a possible path."""
        base_path = str(self.iri).replace('local:', '')
        return f'{base_path}.md'

    @property
    def possible_id(self) -> str:
        """Guess on a possible $id."""
        return str(self.iri).replace('local:', '')


@dataclass
class MultipleFacetsFoundForTypes(DocumentedError):
    """
    Multiple suitable facets found.

    Unable to render the node unambiguously.

      - Node: {self.node}
      - Types:
    {self.formatted_types}
      - Facets found:
    {self.formatted_facets}
    """

    node: Node
    types: List[URIRef]
    facets: List[URIRef]

    @property
    def formatted_facets(self):
        return '\n'.join([
            f'    - {facet}'
            for facet in self.facets
        ])

    @property
    def formatted_types(self):
        return '\n'.join([
            f'    - <{node_type}>'
            for node_type in self.types
        ])


@dataclass
class FacetNotCallable(DocumentedError):
    """
    Python facet not callable.

    !!! error "Cannot import an object or cannot call it."

          - Import path: `{self.path}`
          - Object imported: `{self.facet}`

        The imported Python object is not a callable and thus cannot be used as
        a facet.
    """

    path: str
    facet: object


@dataclass
class FacetNotFound(DocumentedError):
    """
    # Facet not found.

    No way to render the node you asked for.

    - **Node:** `{self.node}` *({self.node_type})*
    - **Environments tried:** `{self.environments}`
    """

    node: Node
    environments: List[NotLiteralNode]
    node_types: List[NotLiteralNode] = field(default_factory=list)

    @property
    def node_type(self) -> str:
        """Node type."""
        node_type = type(self.node).__name__
        if isinstance(self.node, Literal):
            datatype = self.node.datatype
            node_type = f'{node_type}, datatype={datatype}'

        return node_type


@dataclass
class FacetError(DocumentedError):
    """
    Facet rendering failed.

    !!! error "Facet has thrown an unhandled exception"
        - Node: `{self.node}`
        - Facet IRI: `{self.facet_iri}`

        ### Exception

        {self.indented_error}
    """

    node: Node
    facet_iri: URIRef
    error: Exception

    @property
    def indented_error(self):
        """Format the underlying error text."""
        return textwrap.indent(
            str(self.error),
            prefix='    ',
        )


@dataclass
class NotALiteral(DocumentedError):
    """
    Node `{self.node}` is not a literal.

    It is in fact a `{self.node_type}`. `BoolLiteral` facet only supports RDF
    literal objects.
    """

    node: Node

    @property
    def node_type(self):
        """Node type name."""
        return self.node.__class__.__name__



from dataclasses import dataclass, field
from functools import cached_property
from typing import Dict, List, Optional, Type, TypedDict, Iterable

import funcy
from rdflib import ConjunctiveGraph
from rdflib.term import Literal, Node, URIRef

from iolanta.facets.errors import FacetNotFound
from iolanta.models import NotLiteralNode
from iolanta.namespaces import IOLANTA


class FoundRow(TypedDict):
    facet: NotLiteralNode
    environment: NotLiteralNode


@dataclass
class FacetFinder:
    """Engine to find facets for a given node."""

    iolanta: 'iolanta.Iolanta'    # type: ignore
    node: Node
    environments: List[NotLiteralNode]

    @cached_property
    def row_sorter_by_environment(self):
        def _sorter(row) -> int:
            return self.environments.index(row['environment'])

        return _sorter

    def by_datatype(self) -> Iterable[FoundRow]:
        if not isinstance(self.node, Literal):
            return []

        if (data_type := self.node.datatype) is None:
            return []

        rows = self.iolanta.query(   # noqa: WPS462
            """
            SELECT ?environment ?facet WHERE {
                $data_type iolanta:hasDatatypeFacet ?facet .
                ?facet iolanta:supports ?environment .
            }
            """,
            data_type=data_type,
        )

        rows = [row for row in rows if row['environment'] in self.environments]

        return sorted(
            rows,
            key=self.row_sorter_by_environment,
        )

    def by_facet(self) -> List[FoundRow]:
        if isinstance(self.node, Literal):
            return []

        rows = self.iolanta.query(
            '''
            SELECT ?environment ?facet WHERE {
                $node iolanta:facet ?facet .
                ?facet iolanta:supports ?environment .
            }
            ''',
            node=self.node,
        )

        # FIXME This is probably suboptimal, why don't we use `IN environments`?
        rows = [row for row in rows if row['environment'] in self.environments]

        return sorted(
            rows,
            key=self.row_sorter_by_environment,
        )

    def by_instance_facet(self) -> Iterable[FoundRow]:
        rows = self.iolanta.query(
            '''
            SELECT ?environment ?facet WHERE {
                $node a ?class .
                ?class iolanta:hasInstanceFacet ?facet .
                ?facet iolanta:supports ?environment .
            }
            ''',
            node=self.node,
        )

        rows = [row for row in rows if row['environment'] in self.environments]

        return sorted(
            rows,
            key=self.row_sorter_by_environment,
        )

    def by_environment_default_facet(self) -> Iterable[FoundRow]:
        """Find facet based on environment only."""
        graph: ConjunctiveGraph = self.iolanta.graph

        triples = graph.triples(     # type: ignore
            (None, IOLANTA.hasDefaultFacet, None),
        )
        triples = [
            triple
            for triple in triples
            if funcy.first(triple) in self.environments
        ]

        rows = [
            {
                'facet': facet,
                'environment': environment,
            } for environment, _, facet in triples
        ]

        return sorted(
            rows,
            key=self.row_sorter_by_environment,
        )

    def choices(self) -> Iterable[FoundRow]:
        yield from self.by_datatype()
        yield from self.by_facet()
        yield from self.by_instance_facet()
        yield from self.by_environment_default_facet()

    @property
    def facet_and_environment(self) -> FoundRow:
        if choice := funcy.first(self.choices()):
            return choice

        raise FacetNotFound(
            node=self.node,
            environments=self.environments,
            node_types=[],
        )



import pydoc
from typing import Type, cast

from rdflib import URIRef

from iolanta.facets.errors import FacetNotCallable
from iolanta.facets.facet import Facet
from iolanta.resolvers.base import Resolver


class PythonImportResolver(Resolver):
    def __getitem__(self, item: URIRef) -> Type[Facet]:
        url = str(item)

        if not url.startswith('python://'):
            raise Exception(
                'Iolanta only supports facets which are importable Python '
                'callables. The URLs of such facets must start with '
                '`python://`, '
                'which {url} does not comply to.'.format(
                    url=url,
                ),
            )

        # It is impossible to use `yarl` for this operation because it (or,
        # rather, one of upper classes from `urllib` that `yarl` depends
        # upon)
        # will lowercase the URL when parsing it - which means, irreversibly. We
        # have to resort to plain string manipulation.
        import_path = url.replace('python://', '').strip('/')

        facet = cast(Type['iolanta.Facet'], pydoc.locate(import_path))

        if not callable(facet):
            raise FacetNotCallable(
                path=import_path,
                facet=facet,
            )

        return facet



import funcy
from rdflib import URIRef
from textual.containers import Vertical
from textual.widget import Widget
from textual.widgets import Label, Button, Static, ListView, ListItem

from iolanta.facets.facet import Facet


class InstancesGrid(Vertical):
    DEFAULT_CSS = """
    InstancesGrid {
        layout: grid;
        grid-size: 3 2;
    }
    """


class Class(Facet[Widget]):
    def show(self) -> Widget:
        instances = funcy.lpluck(
            'instance',
            self.stored_query('instances.sparql', iri=self.iri),
        )

        return ListView(*[
            ListItem(
                Label(
                    self.render(
                        instance,
                        environments=[URIRef('https://iolanta.tech/cli/link')]
                    ),
                )
            )
            for instance in instances
        ])



from dataclasses import dataclass
from typing import Optional

from documented import DocumentedError
from rdflib import URIRef


@dataclass
class YAMLError(DocumentedError):
    """
    Invalid YAML.

    File: {self.iri}

    {self.error}
    """

    iri: Optional[URIRef]
    error: Exception


@dataclass
class SpaceInProperty(DocumentedError):
    """
    Space in property.

    That impedes JSON-LD parsing.

    Please do not use spaces in property names in JSON or YAML data; use `title`
    or other methods instead.

    Document IRI: {self.iri}
    """

    iri: Optional[URIRef] = None



from dataclasses import dataclass
from typing import Iterable, Optional, TextIO, Type

from rdflib import Literal, URIRef

from iolanta.conversions import url_to_iri
from iolanta.loaders.base import Loader
from iolanta.loaders.errors import IsAContext, ParserNotFound
from iolanta.models import LDContext, LDDocument, Quad
from iolanta.namespaces import IOLANTA
from iolanta.parsers.base import Parser
from iolanta.parsers.dict_parser import DictParser
from iolanta.parsers.json import JSON
from iolanta.parsers.markdown import Markdown
from iolanta.parsers.yaml import YAML


@dataclass(frozen=True)
class DictLoader(Loader[LDDocument]):
    """
    Retrieve Linked Data from a file on local disk.

    Requires a dict of raw JSON-LD data.
    """

    def find_context(self, source: str) -> LDContext:
        raise ValueError('???WTF?')

    def choose_parser_class(self, source: LDDocument) -> Type[Parser]:
        return DictParser(source)

    def as_quad_stream(
        self,
        source: LDDocument,
        root_loader: Loader[LDDocument],
        iri: Optional[URIRef] = None,
        context: Optional[LDContext] = None,
    ) -> Iterable[Quad]:
        """Extract a sequence of quads."""
        yield from DictParser().as_quad_stream(
            raw_data=source,
            iri=iri,
            context=context,
            root_loader=root_loader,
        )

    def as_file(self, source: LDDocument) -> TextIO:
        """Construct a file-like object."""
        raise ValueError('FOO')

    def as_jsonld_document(
        self,
        source: LDDocument,
        iri: Optional[URIRef] = None,
    ) -> LDDocument:
        """As JSON-LD document."""
        return source



from abc import abstractmethod
from typing import Type

from rdflib import URIRef

from iolanta.facets.facet import Facet


class Resolver:
    """Resolve facet IRIs into classes."""
    @abstractmethod
    def __getitem__(self, facet_iri: URIRef) -> Type[Facet]:
        """Find a resolver by IRI."""



from dataclasses import dataclass
from typing import Dict, List, Mapping, Optional, Union

from documented import DocumentedError
from frozendict import frozendict
from pyparsing import ParseException
from rdflib import Graph
from rdflib.plugins.sparql.processor import SPARQLResult
from rdflib.term import Identifier, Node, Variable

from iolanta.cyberspace.processor import GlobalSPARQLProcessor

SelectRow = Mapping[str, Node]


class SelectResult(List[SelectRow]):
    """Result of a SPARQL SELECT."""

    @property
    def first(self) -> Optional[SelectRow]:
        """Return first element of the list."""
        return self[0] if self else None


SPARQLQueryArgument = Optional[Union[Node, str, int, float]]


QueryResult = Union[
    SelectResult,   # SELECT
    Graph,          # CONSTRUCT
    bool,           # ASK
]


def _format_query_bindings(
    bindings: List[Dict[Variable, Identifier]],
) -> SelectResult:
    """
    Format bindings before returning them.

    Converts Variable to str for ease of addressing.
    """
    return SelectResult([
        frozendict({
            str(variable_name): rdf_value
            for variable_name, rdf_value
            in row.items()
        })
        for row in bindings
    ])


@dataclass
class SPARQLParseException(DocumentedError):
    """
    SPARQL query is invalid.

    Error:

    ```
    {self.error}
    ```

    Query:
    ```sparql hl_lines="{self.highlight_code}"
    {self.query}
    ```
    """

    error: ParseException
    query: str

    @property
    def highlight_code(self):
        """Define lines to highlight."""
        return self.error.lineno


@dataclass
class LDFlex:
    """Fluent interface to a semantic graph."""

    graph: Graph

    def query(
        self,
        query_text: str,
        **kwargs: SPARQLQueryArgument,
    ) -> QueryResult:
        """
        Run a SPARQL `SELECT`, `CONSTRUCT`, or `ASK` query.

        Args:
            query_text: The SPARQL text;
            **kwargs: bind variables in the query to values if necessary. For
                example:

                ```python
                ldflex.query(
                    'SELECT ?title WHERE { ?page rdfs:label ?title }',
                    ?page=page_iri,
                )
                ```

        Returns:
            Results of the query:

            - a graph for `CONSTRUCT`,
            - a list of dicts for `SELECT`,
            - or a boolean for `ASK`.
        """
        try:
            sparql_result: SPARQLResult = self.graph.query(
                query_text,
                processor='cyberspace',
                initBindings=kwargs,
            )
        except ParseException as err:
            raise SPARQLParseException(
                error=err,
                query=query_text,
            ) from err

        if sparql_result.askAnswer is not None:
            return sparql_result.askAnswer

        if sparql_result.graph is not None:
            graph: Graph = sparql_result.graph
            for prefix, namespace in self.graph.namespaces():
                graph.bind(prefix, namespace)

            return graph

        return _format_query_bindings(sparql_result.bindings)

    def update(self, sparql_query: str):
        """Apply the given SPARQL INSERT query."""
        self.graph.update(sparql_query)



from dataclasses import dataclass
from pathlib import Path

from documented import DocumentedError
from yarl import URL


@dataclass
class IsAContext(DocumentedError):
    """
    The provided file is a context.

        - Path: {self.path}

    This file is not a piece of data and cannot be loaded into the graph.
    """

    path: URL


@dataclass
class ParserNotFound(DocumentedError):
    """
    Parser not found.

        Path: {self.path}
    """

    path: Path



from dataclasses import dataclass, field

from documented import DocumentedError
from rdflib.term import Node


@dataclass
class InsufficientDataForRender(DocumentedError):
    """
    Insufficient data for rendering {self.node} — will try & download something.
    """

    node: Node
    iolanta: 'iolanta.Iolanta'

    @property
    def is_hopeless(self) -> bool:
        hopeless = self.node in self.iolanta.could_not_retrieve_nodes

        if hopeless:
            self.iolanta.logger.error(
                '%s could not be rendered, we could not retrieve describing '
                'data '
                'from the Web.',
                self.node,
            )

        return hopeless



from textual.app import ReturnType

from iolanta.facets.facet import Facet, FacetOutput
from iolanta.facets.textual_browser.app import IolantaBrowser


class TextualBrowserFacet(Facet[ReturnType | None]):
    """Textual browser."""

    def show(self) -> ReturnType | None:
        app = IolantaBrowser()
        app.iolanta = self.iolanta
        app.iri = self.iri
        app.run()



import functools
import logging
from dataclasses import dataclass, field
from functools import cached_property
from pathlib import Path
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
)

import funcy
import owlrl
from owlrl import OWLRL_Extension
from owlrl.Closure import Core
from rdflib import ConjunctiveGraph, Namespace, URIRef
from rdflib.term import Node

from iolanta import entry_points
from iolanta.errors import InsufficientDataForRender
from iolanta.facets.errors import FacetError
from iolanta.facets.facet import Facet
from iolanta.facets.locator import FacetFinder
from iolanta.loaders import Loader
from iolanta.loaders.base import SourceType
from iolanta.loaders.local_directory import merge_contexts
from iolanta.models import LDContext, NotLiteralNode, Triple, TripleTemplate
from iolanta.namespaces import IOLANTA, LOCAL
from iolanta.parsers.yaml import YAML
from iolanta.plugin import Plugin
from iolanta.resolvers.python_import import PythonImportResolver
from iolanta.shortcuts import construct_root_loader
from iolanta.stack import Stack
from ldflex import LDFlex


@dataclass
class Iolanta:
    """Iolanta is a Semantic web browser."""

    retrieval_directory: Optional[Path] = None
    graph: ConjunctiveGraph = field(
        default_factory=functools.partial(
            ConjunctiveGraph,
            identifier=LOCAL.term('_inference'),
        ),
    )
    force_plugins: List[Type[Plugin]] = field(default_factory=list)

    facet_resolver: Mapping[URIRef, Type[Facet]] = field(
        default_factory=PythonImportResolver,
    )

    logger: logging.Logger = field(
        default_factory=functools.partial(
            logging.getLogger,
            name='iolanta',
        ),
    )

    sources_added_not_yet_inferred: list[SourceType] = field(
        default_factory=list,
        init=False,
        repr=False,
    )

    could_not_retrieve_nodes: Set[Node] = field(
        default_factory=set,
        init=False,
    )

    @cached_property
    def loader(self) -> Loader[SourceType]:
        return construct_root_loader(logger=self.logger)

    @property
    def plugin_classes(self) -> List[Type[Plugin]]:
        """Installed Iolanta plugins."""
        return self.force_plugins or entry_points.plugins('iolanta.plugins')

    @cached_property
    def plugins(self) -> List[Plugin]:
        return [
            plugin_class(iolanta=self)
            for plugin_class in self.plugin_classes
        ]

    @cached_property
    def ldflex(self) -> LDFlex:
        """LDFlex is a wrapper to make SPARQL querying RDF graphs bearable."""
        return LDFlex(self.graph)

    @cached_property
    def namespaces_to_bind(self) -> Dict[str, Namespace]:
        return {
            key: Namespace(value)
            for key, value in self.default_context['@context'].items()
            if (
                isinstance(value, str)
                and not value.startswith('@')
                and not key.startswith('@')
            )
        }

    def add(  # type: ignore
        self,
        source: Any,
        context: Optional[LDContext] = None,
        graph_iri: Optional[URIRef] = None,
    ) -> 'Iolanta':
        """Parse & load information from given URL into the graph."""
        self.logger.info('Adding to graph: %s', source)
        self.sources_added_not_yet_inferred.append(source)

        quads = list(
            self.loader.as_quad_stream(
                source=source,
                iri=graph_iri,
                context=context or self.default_context,
                root_loader=self.loader,
            ),
        )

        self.graph.addN(quads)

        self.bind_namespaces(**self.namespaces_to_bind)

        return self

    def infer(self, closure_class: Type[Core] = OWLRL_Extension) -> 'Iolanta':
        """Apply inference."""
        self.logger.info(
            'Inference: %s started...',
            closure_class.__name__,
        )
        owlrl.DeductiveClosure(closure_class).expand(self.graph)
        self.logger.info('Inference: complete.')

        self.sources_added_not_yet_inferred = []

        return self

    def bind_namespaces(self, **mappings: Namespace) -> 'Iolanta':
        """Bind namespaces."""
        self.graph.bind(prefix='local', namespace=LOCAL)

        for prefix, namespace in mappings.items():
            self.graph.bind(prefix=prefix, namespace=namespace)

        return self

    @property
    def query(self):
        self.maybe_infer()
        return self.ldflex.query

    @cached_property
    def context_paths(self) -> Iterable[Path]:
        directory = Path(__file__).parent / 'data'

        yield directory / 'context.yaml'

        for plugin in self.plugins:
            if path := plugin.context_path:
                yield path

    @cached_property
    def default_context(self) -> LDContext:
        """Construct default context from plugins."""
        context_documents = [
            YAML().as_jsonld_document(path.open('r'))
            for path in self.context_paths
        ]

        for context in context_documents:
            if isinstance(context, list):
                raise ValueError('Context cannot be a list: %s', context)

        return merge_contexts(*context_documents)   # type: ignore

    def add_files_from_plugins(self):
        for plugin in self.plugins:
            self.add(plugin.data_files)

    def __post_init__(self):
        self.add_files_from_plugins()

    def string_to_node(self, name: str | Node) -> Node:
        """
        Parse a string into a node identifier.

        String might be:
          * a full IRI,
          * a qname,
          * a qname with implied `local:` part,
          * or a blank node identifier.
        """
        if isinstance(name, Node):
            return name

        if ':' in name:
            # This is either a full IRI, a qname, or a blank node identifier.
            try:
                # Trying to interpret this as QName.
                return self.graph.namespace_manager.expand_curie(name)
            except ValueError:
                # If it is not a QName then it is an IRI, let's return it.
                return URIRef(name)

        # This string does not include an ":", so we imply `local:`.
        return URIRef(f'local:{name}')

    def render(
        self,
        node: Node,
        environments: List[NotLiteralNode],
    ) -> Tuple[Any, Stack]:
        """Find an Iolanta facet for a node and render it."""
        if not environments:
            raise ValueError(
                f'Please provide at least one environment '
                f'to render {node} against.',
            )

        found = FacetFinder(
            iolanta=self,
            node=node,
            environments=environments,
        ).facet_and_environment

        facet_class = self.facet_resolver[found['facet']]

        facet = facet_class(
            iri=node,
            iolanta=self,
            environment=found['environment'],
        )

        try:
            return facet.show(), facet.stack

        except InsufficientDataForRender:
            raise

        except Exception as err:
            raise FacetError(
                node=node,
                facet_iri=found['facet'],
                error=err,
            ) from err

    def render_all(
        self,
        node: Node,
        environment: NotLiteralNode,
    ) -> Iterable[Any]:
        """Find all possible Iolanta facets for a node and render them."""
        choices = FacetFinder(
            iolanta=self,
            node=node,
            environments=[environment],
        ).choices()

        pairs = [
            (self.facet_resolver[row['facet']], row['environment'])
            for row in choices
        ]

        facet_instances = [
            facet_class(
                iri=node,
                iolanta=self,
                environment=environment,
            )
            for facet_class, environment in pairs
        ]

        for facet in facet_instances:
            try:
                yield facet.show()
            except Exception as err:
                raise FacetError(
                    node=node,
                    facet_iri=None,
                    error=err,
                ) from err

    def retrieve_triple(self, triple_template: TripleTemplate) -> Triple:
        """Retrieve remote data to project directory."""
        for plugin in self.plugins:
            # FIXME Parallelization?
            plugin.retrieve_triple(triple_template)

        if not downloaded_files:
            self.could_not_retrieve_nodes.add(node)

        for path in downloaded_files:
            self.add(path)

        return self

    def maybe_infer(self):
        """
        Apply inference lazily.

        Only run inference if there are new files added after last inference.
        """
        if self.sources_added_not_yet_inferred:
            self.infer()

    def find_triple(
        self,
        triple_template: TripleTemplate,
    ) -> Triple | None:
        """Lightweight procedure to find a triple by template."""
        triples = self.graph.triples(
            (triple_template.subject, triple_template.predicate, triple_template.object),
        )

        raw_triple = funcy.first(triples)
        if raw_triple:
            return Triple(*raw_triple)

        return self.retrieve_triple(triple_template)



from rdflib import URIRef

from iolanta.cli.formatters.node_to_qname import node_to_qname
from iolanta.facets.facet import Facet, FacetOutput
from iolanta.models import ComputedQName, NotLiteralNode


class TextualLinkFacet(Facet[str]):
    def show(self) -> str:
        label = self.render(
            self.iri,
            environments=[URIRef('https://iolanta.tech/env/title')]
        )

        return f'[@click="goto(\'{self.iri}\')"]{label}[/]'



from dataclasses import dataclass, field
from typing import List

from rdflib.term import Node


@dataclass
class Stack:
    node: Node
    facet: 'iolanta.facets.facet.Facet'
    children: List['Stack'] = field(default_factory=list)



import csv
import sys

from classes import typeclass
from rdflib import Graph

from ldflex.ldflex import QueryResult, SelectResult


@typeclass
def csv_print(query_result: QueryResult) -> None:
    """Print as CSV."""


@csv_print.instance(SelectResult)
def _csv_select(select_result: SelectResult):
    if not select_result:
        return

    first_row = select_result[0]
    fieldnames = first_row.keys()

    writer = csv.DictWriter(sys.stdout, fieldnames=fieldnames)
    writer.writeheader()
    writer.writerows(select_result)


@csv_print.instance(Graph)
def _construct_print(graph: Graph):
    if not graph:
        return

    writer = csv.writer(sys.stdout)
    writer.writerow(('subject', 'predicate', 'object'))
    writer.writerows(graph)



from dominate.tags import a, span
from dominate.util import raw

from iolanta.facets.generic.default import DefaultMixin
from iolanta.facets.html.base import HTMLFacet


class Default(DefaultMixin, HTMLFacet):
    """Default renderer."""

    def render_link(self):
        return a(
            self.render_label(),
            href=self.description.url,
            title=self.description.comment,
        )

    def render_label(self):
        label = super().render_label()

        if comment := self.description.comment:
            return span(
                raw(label),
                title=comment,
            )

        return label



from rdflib import URIRef
from textual.app import App, ComposeResult
from textual.containers import ScrollableContainer
from textual.widgets import Header, Footer, Welcome, Button

from iolanta.iolanta import Iolanta
from iolanta.models import NotLiteralNode


class Body(ScrollableContainer):
    """Browser body."""

    def on_mount(self):
        iolanta: Iolanta = self.app.iolanta
        iri: NotLiteralNode = self.app.iri
        self.mount(
            iolanta.render(
                iri,
                [URIRef('https://iolanta.tech/cli/textual')],
            )[0]
        )


class IolantaBrowser(App):
    """Browse Linked Data."""

    iolanta: Iolanta
    iri: NotLiteralNode

    BINDINGS = [
        ('g', 'goto', 'Go to URL'),
        ('s', 'search', 'Search'),
        ('t', 'toggle_dark', 'Toggle Dark Mode'),
        ('q', 'quit', 'Quit'),
    ]

    def compose(self) -> ComposeResult:
        yield Header()
        yield Footer()
        yield Body()

    def action_toggle_dark(self) -> None:
        """An action to toggle dark mode."""
        self.dark = not self.dark

    def action_goto(self, destination: str):
        body = self.query_one(Body)
        body.remove_children()

        self.iri = URIRef(destination)

        iolanta: Iolanta = self.iolanta
        iri: NotLiteralNode = self.iri
        body.mount(
            iolanta.render(
                iri,
                [URIRef('https://iolanta.tech/cli/textual')],
            )[0]
        )



from rdflib import URIRef

from iolanta.facets.facet import Facet


PRIORITIES = [
    'dc_title',
    'schema_title',
    'rdfs_label',
    'foaf_name',
]


class TitleFacet(Facet[str]):
    """Title of an object."""

    def show(self) -> str:
        choices = self.stored_query('title.sparql', iri=self.iri)

        try:
            [row] = choices
        except ValueError:
            raise ValueError(
                f'Not exactly one choice, what do we do?! {choices}',
            )

        for alternative in PRIORITIES:
            if label := row.get(alternative):
                return str(label)

        return self.render(
            self.iri,
            environments=[URIRef('https://iolanta.tech/qname')],
        )



from iolanta.cli.formatters.node_to_qname import node_to_qname
from iolanta.facets.facet import Facet
from iolanta.models import NotLiteralNode, ComputedQName


class QNameFacet(Facet[str]):
    def show(self) -> str:
        qname: ComputedQName | NotLiteralNode = node_to_qname(
            self.iri,
            self.iolanta.graph,
        )

        if isinstance(qname, ComputedQName):
            return f'{qname.namespace_name}:{qname.term}'

        return str(self.iri)



import logging
import sys
import traceback
from functools import cached_property
from pathlib import Path
from typing import Iterable

import funcy
import rdflib
from rdflib import URIRef, DC, SDO, RDFS
from rdflib.term import Node, Literal, BNode
from rich.markdown import Markdown
from textual.app import ComposeResult
from textual.containers import Vertical, VerticalScroll
from textual.widget import Widget
from textual.widgets import Label, Static, DataTable

from iolanta.cli.formatters.node_to_qname import node_to_qname
from iolanta.facets.facet import Facet
from iolanta.models import ComputedQName, NotLiteralNode
from iolanta.namespaces import IOLANTA


class Content(VerticalScroll):
    DEFAULT_CSS = """
    Content {
        layout: vertical;
        overflow-x: hidden;
        overflow-y: auto;
    }
    
    #title {
        padding: 1;
        background: darkslateblue;
    }
    
    #description {
        padding: 1;
    }
        
    #properties {
        padding: 1;
    }
    
    /* FIXME: This one does not work */
    DataTable .datatable--header {
        background: purple;
        color: red;
    }
    """


class TextualDefaultFacet(Facet[Widget]):
    """Default rendering engine."""

    @cached_property
    def grouped_properties(self) -> dict[NotLiteralNode, list[Node]]:
        """Properties of current node & their values."""
        property_rows = self.stored_query(
            'properties.sparql',
            iri=self.iri,
        )

        property_pairs = [
            (row['property'], row['object'])
            for row in property_rows
        ]

        return funcy.group_values(property_pairs)

    @property
    def title(self) -> str:
        """
        Candidates for the page title.

        FIXME: Here, we mutate `grouped_properties` :(

        TODO: Implement locale support.
        """
        choices = [
            title
            for title_property in [DC.title, SDO.title, RDFS.label]
            for title in self.grouped_properties.pop(title_property, [])
            if isinstance(title, Literal)
        ]

        if choices:
            return choices[0].value

        return self.format_clickable_link(self.iri)

    def format_clickable_link(self, iri: NotLiteralNode):
        """Format given IRI as clickable link, if possible."""
        if isinstance(iri, BNode):
            return str(iri)

        qname: ComputedQName | NotLiteralNode = node_to_qname(
            iri,
            self.iolanta.graph,
        )
        if isinstance(qname, ComputedQName):
            namespace_part = (
                f"[@click=goto('{qname.namespace_url}')]{qname.namespace_name}[/]"
            )
            term_part = qname.term if iri == self.iri else (
                f"[@click=goto('{iri}')]{qname.term}[/]"
            )

            return f'{namespace_part}:{term_part}'

        if iri == self.iri:
            return str(iri)

        return f"[@click=goto('{iri}')]{iri}[/]"

    @cached_property
    def description(self) -> str | None:
        """
        Candidates for description.

        FIXME: We mutate `grouped_properties` here.
        """
        choices = [
            description
            for description_property in [
                DC.description,
                SDO.description,
                RDFS.comment,
            ]
            for description in self.grouped_properties.pop(description_property, [])
            if isinstance(description, Literal)
        ]

        try:
            return choices[0].value
        except IndexError:
            return None

    @cached_property
    def properties(self) -> Widget | None:
        if not self.grouped_properties:
            return None

        properties_table = DataTable(show_header=True, show_cursor=False)
        properties_table.add_columns('Property', 'Value')
        properties_table.add_rows([
            (
                self.render(
                    property_iri,
                    environments=[URIRef('https://iolanta.tech/cli/link')]
                ),
                ' · '.join(
                    self.render(
                        property_value,
                        environments=[URIRef('https://iolanta.tech/cli/link')]
                    )
                    for property_value in property_values
                ),
            )
            for property_iri, property_values in self.grouped_properties.items()
        ])

        return properties_table

    def compose(self) -> Iterable[Widget]:
        """Compose widgets."""
        yield Static(
            f'[bold white]{self.title}[/bold white]',
            id='title',
        )

        if self.description:
            yield Label(self.description, id='description')

        sub_facets = list(
            self.render_all(
                self.iri,
                environment=URIRef('https://iolanta.tech/cli/default'),
            ),
        )

        if sub_facets:
            yield from sub_facets

        elif self.properties:
            yield Label('[i]Properties[/i]', id='properties')
            yield self.properties


    def show(self) -> Widget:
        return Content(*self.compose())



        nodes_for_property = [
            (row['subject'], row['object'])
            for row in self.stored_query(
                'nodes-for-property.sparql',
                iri=self.iri,
            )
        ]
        if nodes_for_property:
            rendered_property = self.render(
                self.iri,
                environments=[URIRef('https://iolanta.tech/cli/link')],
            )

            children.append(Label(
                '\n[bold]A few nodes connected with this property[/]\n'
            ))
            nodes_table = DataTable(show_header=False, show_cursor=False)
            nodes_table.add_columns('Subject', 'Property', 'Object')
            nodes_table.add_rows([(
                    self.render(
                        subject_node,
                        environments=[URIRef('https://iolanta.tech/cli/link')],
                    ),
                    rendered_property,
                    self.render(
                        object_node,
                        environments=[URIRef('https://iolanta.tech/cli/link')]
                    ))
                for subject_node, object_node in nodes_for_property
            ])

            children.append(nodes_table)

        if self.grouped_properties:


            children.append(Label('\n[bold]Properties[/]\n'))

        return Vertical(*children)
